"""Module for generating the unit plots, other plot elements like legends, 
and auxiliary functions."""

from . import _profile_renderer
import matplotlib.pyplot as plt
import matplotlib.colors as mc
import numpy as np
import pandas as pd
from decimal import Decimal


def create_default_colors(num_color, color = '#5E5E5E', single = False):
    """Creates default stack colors."""
    if single:
        return [color]*num_color
    cmap_name = 'viridis'  
    cmap = plt.get_cmap(cmap_name)
    listed_cmap = cmap(np.linspace(0, 1, num_color))
    return [mc.to_hex(color) for color in listed_cmap]


class Profile():

    def __init__(self, data):
        """Initializes Profile attributes."""
        # Class attributes
        self.data = data
        self.plotted_already = False
        self.verbose = False

        # Plot attributes
        self.fig = plt.figure(figsize=[8, 4.8])
        self.closed = False
        self.shown = False

        # Data attributes
        self.x = 'mutation'
        self.y = '' 
        self.ytype = ''
        self.group = 'gene'
        self.stack = 'batch'
        self.key_called = False
        self.threshold = 50
        self.struct = []

        # Aesthetic attributes
        self.xticks_fontsize = 6
        self.xticks_rotation = 90
        self.yticks_fontsize = 7
        self.ylabel = ''
        self.ylabel_fontsize = 'medium'
        self.key_fontsize = 8
        #self.group_label = []
        self.group_fontsize = 'small'
        self.group_title = 'Gene'
        self.stack_label = []
        self.stack_fontsize = 'medium'
        self.stack_color =  ''
        self.stack_title = 'Batch'

    def process(self, **process_kwargs):
        """Creates data for plotting and structure.
        
        Returns
        -------
        pandas.DataFrame
            The processed DataFrame for plotting.

        """
        # Checking if the default columns are present or if not, required arguments are provided
        required_data_attributes = ['x', 'group', 'stack']
        default_data_attributes = [self.x, self.group, self.stack]
        default_columns_detected = all(col in self.data.columns.tolist() for col in default_data_attributes)
        required_attributes_detected = all(arg in process_kwargs.keys() for arg in required_data_attributes) #if arg != 'threshold' and arg != 'ytype')
        if not default_columns_detected and not required_attributes_detected:
            raise ValueError(f"Expected columns or arguments not detected.")

        for process_key in process_kwargs.keys():
            setattr(self, process_key, process_kwargs[process_key])

        # Pivoting dataframe to get x counts
        data_pivoted = self.data.copy()
        if self.y == '': # Choosing what to base counts on
            values_for_counting = 'values_for_counting'
            data_pivoted[values_for_counting] = 1
        elif self.y != '':
            values_for_counting = self.y
        data_pivoted = pd.pivot_table(data_pivoted, index=[self.group, self.x], columns=self.stack, values=values_for_counting, aggfunc="sum", fill_value=0).reset_index() 
        data_pivoted.rename_axis(None, axis=1, inplace = True)

        # Applying threshold, keeping only x
        self.stack_names = [stack for stack in data_pivoted.columns if stack not in [self.group, self.x]]
        if len(self.stack_label) == 0: # Assigning stack_names as labels
            self.stack_label = self.stack_names
        data_filtered = data_pivoted.copy()
        for stack in self.stack_names:
            data_filtered[stack] = data_pivoted[stack].apply(lambda x: 0 if x < self.threshold else x)
        
        # Determining whether to normalize or not
        if self.ytype == '':
            if len(self.stack_names) > 1:
                self.ytype = 'weights'
            else:
                self.ytype = 'counts'
        if self.ylabel == '':
            self.ylabel = self.ytype.title()
        if self.ytype == 'weights':
            for stack in self.stack_names: # Using Decimal for precision
                data_filtered[stack] = data_filtered[stack].apply(lambda x: Decimal(str(x)))
                stack_sum = data_filtered[stack].sum()
                data_filtered[stack] = data_filtered[stack] * Decimal('100') / Decimal(str(stack_sum))
                data_filtered[stack] = data_filtered[stack].apply(lambda x: x.quantize(Decimal('0.01'))) 
        
        # Summing x counts across all stacks
        data_filtered['sum'] = data_filtered[self.stack_names].sum(axis=1)
        if self.ytype == 'weights': # Converting Decimal objects back to float
            for col in self.stack_names + ['sum']:
                data_filtered[col] = data_filtered[col].apply(float)
        data_filtered = data_filtered[data_filtered['sum'] > 0]
        
        # Adding keys if provided
        if self.key_called:
            data_with_keys = pd.merge(data_filtered, self.key_data, on=[self.group, self.x], how='outer')
            data_with_keys.fillna(0, inplace=True)
            if data_filtered['sum'].dtype == np.int64:
                numerical_columns = list(data_with_keys.columns)[2:]
                for col in numerical_columns:
                    data_with_keys[col] = data_with_keys[col].astype(np.int64)
            data_with_keys.reset_index(drop=True, inplace=True)
            self.data_for_plotting = data_with_keys
        else:
            self.data_for_plotting = data_filtered
        data_for_plotting = self.data_for_plotting.copy()
        
        # Getting data for calculating structure
        self.group_names = data_for_plotting[self.group].unique().tolist()
        #if len(self.group_label) == 0: # Assigning group_names as labels
        #    self.group_label = self.group_names 
        unique_counts = []
        for group in self.group_names:
            groups_df = data_for_plotting[data_for_plotting[self.group] == group]
            unique_counts.append(len(groups_df[self.x].unique()))
        
        # Getting data for struct
        self.data_for_struct = pd.DataFrame({self.group: self.group_names, 'count': unique_counts})
        self.data_for_struct.sort_values(by='count', ascending=False, inplace=True)
        self.data_for_struct.reset_index(drop=True, inplace=True)
        
        # Returning for vg.stat()
        if self.data_for_plotting.columns.size > 0 and self.data_for_plotting.shape[0] == 0:
            raise ValueError("Plot DataFrame has no rows. Lowering threshold might help.")
        self.data_for_plotting.sort_values(by=[self.group, self.x], inplace=True)
        self.data_for_plotting.reset_index(drop=True, inplace=True)
        if self.verbose:
            print('** Processed data for plotting. **')
        return self.data_for_plotting
        
    def key(self, **key_kwargs):
        """Obtain the keys."""
        self.key_called = True        
        self.key_data = key_kwargs['key_data']
        self.key_label = key_kwargs['key_labels']
        self.key_color = key_kwargs['key_colors']
        if self.verbose:
            print('** Processed key for profile. **')

    def aes(self, **aes_kwargs):
        for aes_key in aes_kwargs.keys():
            setattr(self, aes_key, aes_kwargs[aes_key])
        if self.verbose:
            print('** Processed aesthetics for profile. **')
    
    def struct_method(self, **struct_kwargs):
        """Obtain the structure for the profile (i.e. the groups to include)."""
        row_groups = struct_kwargs['struct_key'].split('/')
        self.struct = [''.join(col.split()).split(',') for col in row_groups]
        if self.verbose:
            print('** Processed struct. **')

    def plot(self):
        """Create the figure."""
        if self.verbose:
            print('** Plotting **')
        
        # Getting structure of the mutation profile grid
        if len(self.struct) == 0:
            self.struct = _profile_renderer.build_struct(self.data_for_struct, self.group)  
        
        # Creating profile grids
        grids_and_axes = _profile_renderer.build_profile_grid(self.struct, self.data_for_struct, 
                                                      self.group, self.key_called)
        label_grid =  grids_and_axes[0]
        legend_grid =  grids_and_axes[1]
        group_title_axes = grids_and_axes[2]
        barplot_axes = grids_and_axes[3]
        heatmap_axes = grids_and_axes[4]
        
        # Gathering aesthetic attributes
        if self.stack_color == '':
            self.stack_color = create_default_colors(len(self.stack_label))
        x_aes = [self.xticks_fontsize, self.xticks_rotation]
        y_aes = [self.yticks_fontsize, self.ylabel, self.ylabel_fontsize]
        group_aes = [self.group_fontsize, self.group_title]
        stack_aes = [self.stack_fontsize, self.stack_label, self.stack_color, self.stack_title]
        if self.key_called:
            key_aes = [self.key_fontsize, self.key_label, self.key_color]
        else:
            key_aes = [[], [], []]
        
        # Building barplots
        group_labels = []
        _profile_renderer.build_profile(group_title_axes, barplot_axes, heatmap_axes,
                                self.data_for_plotting, self.struct, self.group,
                                self.x, self.fig, self.key_called, 
                                key_aes, self.stack_names, stack_aes,
                                group_labels, x_aes, y_aes)
        # Creating figure y-axis label
        _profile_renderer.build_yaxis_label(self.ylabel, label_grid)
        # Creating figure legend
        _profile_renderer.build_legend(legend_grid, stack_aes, group_aes, group_labels)
        plt.tight_layout()

    def show(self):
        """Displays the generated figure."""
        if not self.plotted_already:
            self.plot()
            self.plotted_already = True
        if self.closed:
            show_fig = plt.figure()
            show_manager = show_fig.canvas.manager
            show_manager.canvas.figure = self.fig
            self.fig.set_canvas(show_manager.canvas)
        plt.show()
        self.shown = True
        if self.verbose:
            print('** Showed figure. **')

    def save(self, **save_kwargs):
        """Saves the generated figure or data."""
        if not self.plotted_already:
            self.plot()
            self.plotted_already = True
        file_extension = save_kwargs['fname'].lower().split('.')[-1]
        # Saving CSV
        if file_extension == 'csv':
            if 'fname' in save_kwargs:
                save_kwargs['path_or_buf'] = save_kwargs.pop('fname')
            self.data_for_plotting.to_csv(**save_kwargs)
            if self.verbose:
                print('** Saved data **')
        # Saving figure
        else:
            if self.shown:
                save_fig = plt.figure()
                save_manager = save_fig.canvas.manager
                save_manager.canvas.figure = self.fig
                self.fig.set_canvas(save_manager.canvas)
            plt.savefig(**save_kwargs, bbox_inches='tight')
            plt.close()
            self.closed = True
            if self.verbose:
                print('** Saved figure **')


"""Tests whether processed data provided for plotting is correct."""

from vargram import vargram
import matplotlib.pyplot as plt
import random
import pandas as pd
import re
import tempfile
import os
import shutil
import pytest
from decimal import Decimal


class MyProfileData:

    def __init__(self, key_called, num, ytype):
        self.key_called = key_called
        self.num = num
        self.ytype = ytype

    def create_output(self):
        self.output = pd.DataFrame(
            {
                "gene":self.generate_genes(self.num),
                "mutation":self.generate_mutations(self.num)
            }
        )
        # Adding batch counts
        self.num_batches = 2
        self.key_only_ind = random.sample(range(0,  self.num), round(self.num/3))
        for i in range(self.num_batches):
            batch_counts = self.generate_batch_counts(self.num, threshold=50,max_counts=100)
            if self.key_called: # Zero counts in all batches for certain mutations that are only present for the key lineages
                batch_counts = [count if ind not in self.key_only_ind else 0 for ind, count in enumerate(batch_counts)]                
            self.output[f'batch_{i+1}'] = batch_counts.copy()
        # Adding total counts per gene-mutation pair
        self.output['sum'] = self.output.iloc[:, 2:].sum(axis=1)
        # Adding keys
        if  self.key_called:
            self._add_keys_to_output()
        else:
            self.output = self.output[self.output['sum'] != 0]
        # Reordering
        self.output.sort_values(by=['gene', 'mutation'], inplace=True)
        self.output.reset_index(drop=True, inplace=True)
        if self.ytype == 'weights':
            return self.normalized_output()
        else:
            return self.output
    
    def _add_keys_to_output(self):
        # Adding keys
        num_keys = 2
        # Dividing the pure key mutations among the keys
        floor = len(self.key_only_ind) // num_keys
        remainder = len(self.key_only_ind) % num_keys
        num_pure_key_mutations = [floor]*num_keys
        num_pure_key_mutations[-1] = num_pure_key_mutations[-1] + remainder
        for i in range(num_keys):
            # Setting pure key mutations to 0 for this key (which makes it 1 for other keys)
            this_key_ind = self.key_only_ind[:num_pure_key_mutations[i]]
            self.output[f'key_{i+1}'] = [1 if ind not in this_key_ind else 0 for ind in range(self.num)]
            # Removing sampled pure key indices
            self.key_only_ind = self.key_only_ind[num_pure_key_mutations[i]:]
    
    def normalized_output(self):
        self.normalized = self.output.copy()
        numerical_columns = self.normalized.columns.tolist()[2:]
        batches = numerical_columns[:self.num_batches]
        for batch in batches: # Using Decimal for precision
            self.normalized[batch] = self.normalized[batch].apply(lambda x: Decimal(str(x)))
            batch_sum = self.normalized[batch].sum()
            self.normalized[batch] = self.normalized[batch] * Decimal('100') / Decimal(str(batch_sum))
            self.normalized[batch] = self.normalized[batch].apply(lambda x: x.quantize(Decimal('0.01')))        
        self.normalized['sum'] = self.normalized[batches].sum(axis=1)
        for col in numerical_columns:
            self.normalized[col] = self.normalized[col].apply(float)
        return self.normalized

    def create_input(self):
        # Getting batch names
        column_names = list(self.output.columns)
        sum_index = column_names.index("sum")
        batch_names = column_names[2:sum_index]            
        rows = []
        for _, row in self.output.iterrows():
            gene = row['gene']
            mutation = row['mutation']            
            # Iterating through batch columns
            for batch in batch_names:
                count = row[batch]                
                # Appending the gene and mutation pair to the rows list 'count' times
                for _ in range(count):
                    rows.append([batch, gene, mutation])
        # Creating new DataFrame from the rows
        self.input = pd.DataFrame(rows, columns=['batch', 'gene', 'mutation'])
        mutation_positions = self.input['mutation'].apply(self.extract_position)
        self.input.insert(loc=2, column='position', value=mutation_positions)
        self.input.sort_values(by='position', inplace=True)
        return self.input
    
    def create_keys(self):
        if not self.key_called:
            raise NotImplementedError("Keys cannot be created if key_called is false.")        
        # Getting key names
        column_names = list(self.output.columns)
        sum_index = column_names.index("sum")
        key_names = column_names[sum_index+1:]
        keys = []
        for key_name in key_names:
            keys.append(self.output[self.output[key_name] == 1])
        return keys
    
    def generate_mutations(self, num_mutations, 
                           amino_acids='ACDEFGHIKLMNPQRSTVWY', 
                           max_position=10000):
        unique_positions = random.sample(range(1, max_position + 1), num_mutations)
        sorted_positions = sorted(unique_positions.copy()) 
        # All mutations are guaranteed to be unique
        mutations = []
        for pos in unique_positions:
            mutation_type = random.randint(0,3)
            if mutation_type <= 1: # substitution
                mutations.append(f'{random.choice(amino_acids)}{pos}{random.choice(amino_acids)}')
            elif mutation_type == 2: # deletion
                pos_ind = sorted_positions.index(pos)
                if pos == max(sorted_positions): 
                    end_position = random.randint(1,10)
                else: # ensuring end_position does not include any unique_position
                    end_positions = [epos for epos in range(sorted_positions[pos_ind+1] - pos + 1) if not any(pos < other_pos < pos + epos for other_pos in sorted_positions)]
                    end_position = pos+random.sample(end_positions, 1)[0]
                mutations.append(f'{pos}-{end_position}')
            elif mutation_type == 3: # insertion
                insertion_length = random.randint(1, 10)
                mutations.append(f"{pos}:{''.join(random.choices(amino_acids, k = insertion_length))}") 
        return mutations

    def generate_genes(self, num_genes, num_unique_genes = 10):        
        unique_genes = [f'GENE_{i:02}' for i in range(1, num_unique_genes + 1)]
        return [random.choice(unique_genes) for _ in range(num_genes)]

    def generate_batch_counts(self, num_counts, max_counts = 100, threshold = 50):
        counts = [random.randint(threshold, max_counts) for _ in range(num_counts)]
        counts = [count if random.randint(0,1) == 1 else 0 for count in counts]
        return counts

    def extract_position(self, mutation):
        match = re.search(r'\d+\.?\d*', mutation)
        return float(match.group())


@pytest.fixture(params=[(False, 0, 'counts'), (False, 10, 'counts'),
                        (False, 0, 'weights'), (False, 10, 'weights'),
                        (True, 0, 'counts'), (True, 10, 'counts'),
                        (True, 0, 'weights'), (True, 10, 'weights')])
def profile_data(request):
    key_called, threshold, ytype = request.param
    num = random.randint(30,100)
    # Creating test outputs and inputs
    mbd = MyProfileData(key_called=key_called, num=num, ytype=ytype)
    output = mbd.create_output()
    input = mbd.create_input()
    # Getting vargram output
    vg = vargram(data=input)
    vg.profile(threshold=threshold,ytype=ytype)
    if key_called == True:
        keys = mbd.create_keys()
        try:
            vargram_test_dir = tempfile.mkdtemp(prefix="vargram_test_dir")
            for i, key in enumerate(keys):
                key_path = os.path.join(vargram_test_dir, f'key_{i+1}.csv')
                key.to_csv(key_path, index=False)
                key.head(5)
                vg.key(key_path)
        finally:
            shutil.rmtree(vargram_test_dir)
    yield {"vg":vg, "output":output}
    plt.close()


class TestProfileData:

    def test_returned_profile_data(self, profile_data):
        """Returned profile data should be equal to saved data. 
        """ 
        vg = profile_data["vg"]
        result = vg.stat()
        try:
            vargram_test_dir = tempfile.mkdtemp(prefix="vargram_test_dir")
            saved_file_path = os.path.join(vargram_test_dir, 'saved_profile_data.csv')
            vg.save(saved_file_path, index=False)
            expected = pd.read_csv(saved_file_path)        
            assert result.equals(expected)
        finally:
            shutil.rmtree(vargram_test_dir)

    def test_batch_key_sums(self, profile_data):
        """The total sum of all batch and key columns should not be zero.
        """
        vg = profile_data["vg"]
        column_sums = vg.stat().sum(numeric_only=True, axis=1)
        zero_column_sums = column_sums.abs() < 1e-10
        result = zero_column_sums.any()
        expected = False
        assert result == expected
    
    def test_number_mutations(self, profile_data):
        """The length of the dataframe should be equal to the number of unique mutations.
        This is true for the test input created.
        """
        vg = profile_data["vg"]
        output = vg.stat()
        result = len(output['mutation'])
        expected = len(output['mutation'].unique())
        assert result == expected

    def test_profile_data(self, profile_data):
        """The returned profile data should be equal to expected profile data.
        """
        vg = profile_data["vg"]
        expected = profile_data["output"]
        result = vg.stat()
        assert result.equals(expected) 


"""Main module for generating VARGRAM figures and summary statistics."""

from .wranglers._nextclade import nextclade
from .wranglers import _nextclade_utils
from .plots._profile import Profile
import pandas as pd
import os


class vargram:

    def __init__(self, **vargram_kwargs):
        """Initializes the VARGRAM object, getting data 
        and joining it with metadata (if provided).

        Parameters
        ----------
        seq : str
            FASTA file path of the sequences.
        ref : str
            FASTA file path of the reference sequence.
        gene : str
            GFF3 file path of the genome annotation.
        data : str or pandas.DataFrame
            The data to be plotted.
        metadata : string or pandas.DataFrame
            The metadata to be joined with the data.
        join : str or list
            Provides the column to do an outer merge on. 
            If data and metadata do not share a column, 
            the two column names which share the values of interest may be provided.
        
        Returns
        -------
        None

        Raises
        ------
        ValueError
            If neither 'seq' nor 'data' is provided.
            If 'metadata' is provided but 'join' is not.
            If the resulting '_data' attribute is empty.

        """
        # Defining initial values
        self._initialize_variables()

        # Getting data
        if 'seq' in vargram_kwargs.keys(): # Get Nextclade output
            self._nextclade_seqname = 'seqName'
            self._aasub = 'aaSubstitutions'
            self._aadel = 'aaDeletions'
            self._aains = 'aaInsertions'
            self._nextclade_seqname = 'seqName'
            nextclade_kwargs = {key: vargram_kwargs[key] for key in ['seq', 'ref', 'gene'] if key in vargram_kwargs.keys()}
            self._data = nextclade(**nextclade_kwargs)
            self._data = _nextclade_utils.process_nextclade(self._data)
            self._nextclade_called = True
        elif 'data' not in vargram_kwargs.keys():
            missing_data_error=("Missing data. Either provide the FASTA files with the 'seq' "
                                "and 'ref' arguments or provide a dataframe/path with 'data'.")
            raise ValueError(missing_data_error)
        else:
            if isinstance(vargram_kwargs['data'], str): # Data is a path
                ext = os.path.splitext(vargram_kwargs['data'])[1]
                if 'nextclade' in vargram_kwargs.keys() and vargram_kwargs['nextclade'] == True:
                    delimiter = ';'
                elif ext == '.csv':
                    delimiter = ','
                elif ext == '.tsv':
                    delimiter = '\t'
                read_data = pd.read_csv(vargram_kwargs['data'], delimiter=delimiter)
            else:
                read_data = vargram_kwargs['data'] # Data is a dataframe
            # Processing if provided CSV is Nextclade output
            if 'nextclade' in vargram_kwargs.keys() and vargram_kwargs['nextclade'] == True:
                if 'batch' not in read_data.columns:
                    read_data.insert(0, 'batch', 'my_batch')
                read_data.sort_values(by=['batch', 'seqName'], inplace=True)
                read_data.reset_index(drop=True, inplace=True)
                self._data = _nextclade_utils.process_nextclade(read_data)
            else:
                self._data = read_data

        # Getting metadata and joining it with data
        if 'meta' in vargram_kwargs.keys():
            # Getting columns to merge on 
            if 'join' not in vargram_kwargs.keys():
                missing_join_error=("Missing 'join' argument. Provide column name(s) " 
                                    "to (outer) merge data and metadata on.")
                raise ValueError(missing_join_error)
            elif isinstance(vargram_kwargs['join'], str): 
                join = [vargram_kwargs['join']]
            else:
                join = vargram_kwargs['join']
            metadata = vargram_kwargs['meta']
            if len(join) != 1:
                metadata.rename(columns={join[1]: join[0]}, inplace=True)
                self._data = pd.merge(self._data, metadata, how='outer', on=join[0])
            elif 'seq' in vargram_kwargs.keys(): 
                # If 'seq' is provided, automatically join on nextclade_sequence_name
                metadata.rename(columns={join[0]: self._nextclade_seqname}, inplace=True)
                self._data = pd.merge(self._data, metadata, 
                                      how='outer', on=self._nextclade_seqname)
            else: 
                self._data = pd.merge(self._data, metadata, how='outer', on=join)                
        if self._data.empty:
            raise ValueError("Data is empty.")
    
    def _initialize_variables(self):
        """Sets initial values of attributes."""
        self._methods_called = []
        self._methods_kwargs = []
        self._plots = ['profile'] # These are 'plot methods'
        self._terminals = ['show', 'save', 'stat'] # These are the 'terminal methods'
        self._generate_plot = False
        self._plots_indices = []
        self._clean_keys()
        self._shown = False
        self._saved = False
    
    def _clean_keys(self):
        """Flushes key variables clean."""
        self._master_key_data = pd.DataFrame()
        self._nkeys = 0
        self._key_labels = []
        self._key_colors = []
    
    def _generate(self):
        """Runs all called methods in correct order"""
        # Getting group of methods called since last called plot
        latest_method_calls = self._methods_called[self._latest_plot_index:]
        latest_method_kwargs = self._methods_kwargs[self._latest_plot_index:]

        # Determining whether to generate plots or not
        if not self._generate_plot:
            getattr(self, latest_method_calls[-1])(**self._methods_kwargs[-1])
            return
        
        # Creating plot object instance
        plot_class = latest_method_calls[0][1:].title() 
        plot_object = globals()[plot_class]
        self._plot_instance = plot_object(self._data)

        # Rearranging so that auxiliary methods are run before plot and save/show methods
        latest_method_calls.append(latest_method_calls[-1])
        latest_method_calls[-2] = latest_method_calls[0]
        latest_method_calls.pop(0)
        latest_method_kwargs.append(latest_method_kwargs[-1])
        latest_method_kwargs[-2] = latest_method_kwargs[0]
        latest_method_kwargs.pop(0)

        # Generating figure based on most recent plot called
        run_key = True
        for i, method in enumerate(latest_method_calls):
            # Running each method since the most recent plot called
            if run_key or method != '_key': # Ensuring that the internal key method is only called once
                if method == '_key':
                    run_key = False
                if method == '_profile':
                    self._plot_data = getattr(self, method)(**latest_method_kwargs[i])
                else:
                    getattr(self, method)(**latest_method_kwargs[i])
            else:
                continue
        self._generate_plot = False
    
    def _show(self, **_show_kwargs): 
        """Show generated figure"""
        getattr(self._plot_instance, 'show')()
    
    def _save(self, **_save_kwargs):
        """Save generated figure or data"""
        getattr(self._plot_instance, 'save')(**_save_kwargs)
    
    def _stat(self, **_stat_kwargs):
        """Kept for consistency with _generate() flow."""
        pass
    
    def _profile(self, **_profile_kwargs):
        """Process data for plotting"""
        return getattr(self._plot_instance, 'process')(**_profile_kwargs)

    def _key(self, **_key_kwargs):
        """Process key data for plotting"""
        getattr(self._plot_instance, 'key')(key_data=self._master_key_data, 
                                            key_labels=self._key_labels, 
                                            key_colors=self._key_colors)
    
    def _struct(self, **_struct_kwargs):
        """Modify aesthetic attributes"""
        getattr(self._plot_instance, 'struct_method')(**_struct_kwargs)

    def _aes(self, **_aes_kwargs):
        """Modify aesthetic attributes"""
        getattr(self._plot_instance, 'aes')(**_aes_kwargs)
    
    def stat(self):
        """Wrapper for stat method
        
        Returns
        -------
        None

        """
        self._methods_called.append('_stat')
        self._methods_kwargs.append({'empty_string':''}) 
        # The unused empty string argument is so as to be able to maintain 
        # length of methods and methods_kwargs the same
        self._generate()
        return self._plot_data
        
    def show(self): 
        """Wrapper for show method"""
        self._shown = True
        self._methods_called.append('_show')
        self._methods_kwargs.append({'empty_string':''}) 
        # The unused empty string argument is so as to be able to maintain
        # length of methods and methods_kwargs the same
        self._generate()

    def save(self, fname, **save_kwargs):
        """Wrapper for save method"""
        self._saved = True
        save_kwargs['fname'] = fname
        self._methods_called.append('_save')
        self._methods_kwargs.append(save_kwargs)
        self._generate()

    def profile(self, **profile_kwargs):
        """Captures profile method arguments
        
        Parameters
        ----------
        x : str, default:'mutation'
            The x-axis variable (column name of data).
        y : str
            The y-axis variable (column name of data).
        ytype : {'counts', 'weights'}
            The type of calculation for the y-axis.
            'counts' indicate raw counts.
            'weights' indicate normalization by total count.
        group : str, default:'gene'
            The column name for the groups.
        stack : str, default:'batch'
            The column name for the stacks.
        threshold : int, default:50
            The minimum number of occurences of an x value to be included.

        Returns
        -------
        None

        """
        self._methods_called.append('_profile')
        self._methods_kwargs.append(profile_kwargs)
        self._latest_plot_index = len(self._methods_called) - 1
        self._generate_plot = True
        self._clean_keys()

    def key(self, key_data, **key_kwargs):
        """Joins all key data
        
        Parameters
        ----------
        key_data : str
            File path of key lineage
        x : str, default:'mutation'
            The x-axis variable (column name of data).
        group : str, default:'gene'
            The column name for the groups.
        label : str
            The name of the key lineage.
        
        Returns
        -------
        None

        """
        self._methods_called.append('_key')
        self._methods_kwargs.append({'empty_string':''}) 
        # The unused empty string argument is so as to be able to maintain 
        # length of methods and methods_kwargs the same

        self._nkeys += 1
        # Reading data
        if isinstance(key_data, str):
            key_read = self._read_comma_or_tab(key_data)
        if isinstance(key_data, pd.DataFrame):
            key_read = key_data.copy()

        # Getting x to read
        if 'x' in key_kwargs.keys():
            key_x = key_kwargs['x']
        else:
            key_x = 'mutation'

        # Getting group to read
        if 'group' in key_kwargs.keys():
            key_group = key_kwargs['group']
        else:
            key_group = 'gene'

        # Just keeping the 'x' and 'group' columns
        key_df = key_read[[key_group, key_x]].copy()

        # Getting name of key
        if 'label' in key_kwargs:
            key_label = key_kwargs['label']
        elif isinstance(key_data, str):
            key_label = os.path.basename(key_data)
            key_label = os.path.splitext(key_label)[0]
        else:
            key_label = f'key_{self._nkeys}'
        self._key_labels.append(key_label)

        # Getting color of key
        if 'color' in key_kwargs:
            color = key_kwargs['color']
        else:
            color = '#5E5E5E'
        self._key_colors.append(color)
        
        # Gathering in one master dataframe
        if self._nkeys > 1:
            key_df[key_label] = 1
            self._master_key_data = pd.merge(self._master_key_data, key_df, 
                                             on=[key_group, key_x], how='outer')
            self._master_key_data.fillna(0, inplace=True)
            self._master_key_data.reset_index(drop=True, inplace=True)
        else:
            self._master_key_data = pd.DataFrame()
            self._master_key_data[key_group] = key_df[key_group]
            self._master_key_data[key_x] = key_df[key_x]
            self._master_key_data[key_label] = 1

    def struct(self, struct_arg):
        """Retains only groups that are provided.
        
        Parameters
        ----------
        struct_arg : str
            A formatted string indicating the groups to retain per row.
            Commas separate group names and forward slashes separate rows.

        Returns
        -------
        None

        """
        self._methods_called.append('_struct')
        self._methods_kwargs.append({'struct_key':struct_arg})

    def aes(self, **aes_kwargs):
        """Captures modification of aesthetic attributes.
        
        Parameters
        ----------
        stack_label : str
            Name(s) of the stack(s).
        stack_title : str
            Legend title for the stack(s).
        stack_fontsize : float
            The fontsize.
        stack_color : str or list
            Color(s) of the stack(s).
        group_label : str or list
            Name(s) of the group(s).
        group_title : str
            Legend title for the groups.
        group_fontsize : float
            The fontsize.
        xticks_fontsize : float
            The fontsize of the x-axis ticks.
        xticks_rotation : float
            The degree of rotation of the x-axis ticks.
        ylabel : str
            The y-axis label.
        ylabel_fontsize : float
            The fontsize of the y-axis label.
        key_fontsize : float
            The fontsize of the heatmap (key) row labels.

        Returns
        -------
        None

        """
        self._methods_called.append('_aes')
        self._methods_kwargs.append(aes_kwargs)

    def _read_comma_or_tab(self, file):
        """Reads a CSV or TSV file.

        Parameters
        ----------
        file : str
            The file path to be read.
        
        Returns
        -------
        pandas.DataFrame
            The DataFrame of the file.
        
        Raises
        ------
        ValueError
            If the given file is not a CSV or TSV file.

        """
        root, extension = os.path.splitext(file)
        if extension == '.csv':
            return pd.read_csv(file)
        elif extension == '.tsv':
            return pd.read_csv(file, delimiter='\t')
        else:
            raise ValueError("Incorrect file format. Expecting CSV or TSV file.")        